from datetime import timedelta

from fastapi import Depends, FastAPI, HTTPException, status
from fastapi.security import OAuth2PasswordRequestForm

from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from sqlalchemy.orm import selectinload, joinedload

from .models.authentication import (
  User, UserRead,
  Role, RoleRead,
  UserRoleLink
)
from .models.category import (
  CategoryGroup, CategoryModelType,
  Category, CategoryRead
)
from .database import get_db_session, pandas_query
from .security import (
    User, Token,
    authenticate_user, get_current_active_user, create_access_token,
    )
from .config import conf

api = FastAPI()


@api.get("/nothing")
async def get_nothing() -> str:
    return ''

@api.post("/token", response_model=Token)
async def login_for_access_token(form_data: OAuth2PasswordRequestForm = Depends()):
    user = await authenticate_user(form_data.username, form_data.password)
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    access_token_expires = timedelta(
        minutes=conf.security['access_token_expire_minutes'])
    access_token = create_access_token(
        data={"sub": user.username},
        expires_delta=access_token_expires)
    return {"access_token": access_token, "token_type": "bearer"}

@api.get("/users")
async def get_users(
    *, db_session: AsyncSession = Depends(get_db_session)
    ) -> list[UserRead]:
    query = select(User).options(selectinload(User.roles))
    data = await db_session.exec(query)
    return data.all()

@api.get("/roles")
async def get_roles(
    *, db_session: AsyncSession = Depends(get_db_session)
    ) -> list[RoleRead]:
    query = select(Role).options(selectinload(Role.users))
    data = await db_session.exec(query)
    return data.all()

@api.get("/categories")
async def get_categories(
    *, db_session: AsyncSession = Depends(get_db_session)
    ) -> list[CategoryRead]:
    query = select(Category)
    data = await db_session.exec(query)
    return data.all()


@api.get("/categories_p")
async def get_categories_p(
    *, db_session: AsyncSession = Depends(get_db_session)
    ) -> list[CategoryRead]:
    query = select(Category)
    df = await db_session.run_sync(pandas_query, query)
    return df.to_dict(orient="records")

# @api.get("/user-role")
# async def get_user_role_relation(
#     *, db_session: AsyncSession = Depends(get_db_session)
#     ) -> list[UserRoleLink]:
#     roles = await db_session.exec(select(UserRoleLink))
#     return roles.all()